
library(shiny)
library(shinythemes)

ui <- navbarPage(theme=shinytheme("flatly"),
                 title="",
                 tabPanel("Overview of PSAMM",
                          h4("PSAMM is an open source software that is designed for the curation and analysis of metabolic models. 
                             It supports model version tracking, model annotation, data integration, data parsing and formatting, consistency checking, 
                             automatic gap filling, and model simulations. The following figure is the overview workflow in PSAMM:"),
                          tags$img(src="psamm_image.png", heigh = 800, width = 800),
                          p("(Figure taken from Jon Lund Steffensen et al, 2016)"),
                          p("psamm tutorial can be found here: https://psamm.readthedocs.io/en/stable/index.html")),
                 
                 tabPanel("Usage of this app",
                          h3("In this app, two import functions and six analysis functions are included."),
                          h4("- 2 import functions:"),
                          p("convert SBML or JSON format metabolic model to YAML format"),
                          h4("- 6 analysis functions:"),
                          p(paste("masscheck:", "Check whether the model is mass consistent", sep = "\t")),
                          p(paste("chargecheck:", "Check whether compound charge is balanced in a reaction", sep = '\t')),
                          p(paste("formulacheck:", "Check whether reactions in the model are elementally balanced", sep = '\t')),
                          p(paste("fba:", "Run flux balance analysis on the model", sep = '\t')),
                          p(paste("fva:", "Run flux variablity analysis on the model", sep = '\t')),
                          p(paste("search:", "Search for reactions and compounds in the model", sep = '\t')),
                          p(" "),
                          h4("Notes:"), 
                          p("1. In order to run fba and fva successfully,the objective reaction (bioomass reaction) need to be specified"),
                          p("2. If charge or formula of compounds are not defined in the model, then chargecheck or formulacheck can't export 
                            reliable results."),
                          
                          h4("In 'Curation and analysis functions in PSAMM' part, the uploaded model file 
                             should define the reactions, compounds and medium(exchange.yaml) used in the mdoel. It looks like the following:"),
                          
                          tags$img(src="model_example.png", heigh = 500, width = 500),
                          p("In the model file example above, the '<PATH>' should be absolute path 
                            that locates to other files included in the model")),
                 
                 tabPanel("Import functions in PSAMM",
                          headerPanel(
                            h4("PSAMM requires YAML formatted metabolic model as input for further analysis.
                               This format allows for a human readable representation of the model components and 
                               allows for enhanced customization with respect to the organization of the metabolic 
                               model, thus is more accessible for both modeler and experimental biologists. Here PSAMM 
                               provides several functions to convert popular model format to YAML format.")),
                          sidebarLayout(
                            sidebarPanel(
                              h3("Import functions"),
                              p("Convert other formats of metabolic model (e.g. SBML, JSON) to YAML format"),
                              fileInput("raw_model", h5("Choose model file (SBML or JSON)"), accept = c(".xml", ".json"),  multiple = FALSE),
                              radioButtons("current_format", label = h5("Format of inout model file"),
                                           choices = c("SBML", "JSON")),
                              # downloadButton("downloadData", "Download")
                              textInput("dest", label = h5("Abolute path of directory that converted YAML model will be stored"), value='')
                            ),
                            
                            # Main panel for displaying outputs ----
                            mainPanel(
                              uiOutput("converted_model")
                            )
                          )
                            ),
                 
                 tabPanel("Curation and analysis functions in PSAMM",
                          sidebarLayout(
                            sidebarPanel(
                              h3("Curation & Analysis functions"),
                              p("Functions supported by PSAMM that are used for model curation and analysis. 
                                These function can be used to correct common errors and ensure that metabolic reconstructions are accurate 
                                representations of the metabolism of an organism, as well as perform various 
                                simulations of growth with metabolic models. "),
                              fileInput("model", h4("Choose model file (YAML format model)"), accept = ".yaml",  multiple = FALSE),
                              # shinyDirButton('Model_directory', 'Folder select', 'Please select a folder'),
                              selectInput(inputId = "Function_choose", label = h4("Select functions you want to apply on your model"),
                                          choices = c('masscheck', 'chargecheck', 'formulacheck', 'fba', 
                                                      'fva', 'search')),
                              radioButtons("material", label = h4("Search for reactions and compounds in the model(only used for 'search') "),
                                           choices = c("reaction", "compound")),
                              textInput("ID", label = h4("Enter reaction or compound ID (only used for 'search' function)"), value = "Enter reaction or compound ID...")  
                              
                              ),
                            
                            
                            mainPanel(
                              tableOutput("result"),
                              tableOutput("detail")
                            )
                              )
                            )
                 
                 )

# step 2: Define server logic required to draw a histogram ----
server <- function(input, output, session) {
  Sys.setenv(PATH=paste("./python_env/bin",
                        Sys.getenv("PATH"), sep=.Platform$path.sep), VIRTUAL_ENV = "./python_env/")
  Sys.unsetenv("PYTHONHOME")
  
  dataInput <- reactive({
    command_string = input$Function_choose
    if (is.null(input$model)){
      final_command = paste("psamm-model --model ./E_coli_yaml/model.yaml", command_string, sep=' ')
    } else{
      model_file = input$model
      model_path = model_file$datapath
      final_command = paste("psamm-model --model", model_path, command_string, sep = ' ')
      return(final_command)
    }
  })
  
  data2 <- reactive({
    source <- input$raw_model$datapath
    target <- input$dest
    if (is.null(input$raw_model)){
      import_command <- "psamm-import sbml --source ./E_coli_sbml/ecoli_core_model.xml --dest ./sbml_model_import_result"
    } else{
      import_command <- paste("psamm-import", input$current_format, "--source", source, 
                              "--dest", target, sep=' ')
    }
    return(import_command)
  })
  
  output$converted_model <- renderUI(
    system(data2(), intern = TRUE)
  )
 
  
  output$result <- renderTable({
    if (input$Function_choose!="search") {
      a <- system(paste(dataInput(), " 2>&1"), intern = TRUE)
      if(input$Function_choose=="masscheck"){
        grep('Consistent compounds:', a, value=TRUE)
      } else {
        if (input$Function_choose=="formulacheck"){
          grep('reactions', a, value=TRUE)
          } else{
            if (input$Function_choose=="chargecheck"){
              grep('INFO:', a, value=TRUE)
              } else {
                if (input$Function_choose=="fba"){
                  grep('flux:', a, value=TRUE)
                  } else {
                    if (input$Function_choose=="fva"){
                      grep('objective', a, value=TRUE)
                      }
                    }
                }
            }
        }
      }
  })
  
  
  # output$result <- renderTable({
  #   # a <- read.table(text=system(paste(dataInput()), intern = TRUE), sep="\t", header=FALSE)
  #   if (input$Function_choose!="search") {
  #     a <- read.table(text=system(paste(dataInput(), " 2>&1"), intern = TRUE), sep="\t", header=FALSE)
  #     tail(a,3)
  #     if(length(system(paste(dataInput()), intern = TRUE)) != 0){b}
  #     else{"All reactions in this model is charge balanced."}
  #     # b <- system(paste(dataInput(), " 2"), intern = TRUE)
  #     # a <- read.table(text=system(paste(dataInput()), intern = TRUE), sep="\t", header=FALSE)
  #     # if (is.null(read.table(text=system(paste(dataInput()), intern = TRUE), sep="\t", header=FALSE))){
  #     #   a <- read.table(text=system(paste(dataInput(), " 2>&1"), intern = TRUE), sep="\t", header=FALSE)
  #     # }
  #     
  #     # psate("Results:", a, sep = "\n")
  #   } else {if(input$Function_choose=="search") {
  #     a <- read.table(text=system(paste(dataInput(), input$material, '--id', input$ID, sep=' '), intern = TRUE), sep="\t", header=FALSE)
  #     colnames(a) <- c("Result:")
  #   } else {
  #     a <- read.table(text=system(paste(dataInput()), intern = TRUE), sep="\t", header=FALSE)
  #     if(input$Function_choose=="masscheck"){
  #       colnames(a) <- c("compound_id", "assigned_mass", "compound_name")
  #       a
  #       # b <- read.table(text=system(paste(dataInput(), " 2>&1"), intern = TRUE), sep="\t", header=FALSE)
  #       # colnames(b) <- c("Result summary")
  #       # b
  #     } else{if(input$Function_choose=="fba")
  #       colnames(a) <- c("compound_id", "assigned_mass", "compound_name")
  #     a
  #     }
  #   }}})
      
  output$detail <- renderTable({
    # a <- read.table(text=system(paste(dataInput()), intern = TRUE), sep="\t", header=FALSE)
    if (input$Function_choose=="search") {
      a <- read.table(text=system(paste(dataInput(), input$material, '--id', input$ID, sep=' '), intern = TRUE), sep="\t", header=FALSE)
      colnames(a) <- c("Result:")
      a} else {
        tmp = system(paste(dataInput()), intern = TRUE)
        if (length(tmp) != 0){
          a <- read.table(text=system(paste(dataInput()), intern = TRUE), sep="\t", header=FALSE)
          if(input$Function_choose=="masscheck"){
            colnames(a) <- c("compound_id", "assigned_mass", "compound_name")
            a} else {
              if (input$Function_choose=="formulacheck"){
                colnames(a) <- c("reaction_id", "formula_left", "formula_right", "formula_differnce")
                a} else{
                  if (input$Function_choose=="chargecheck"){
                    colnames(a) <- c("reaction_id", "charge_difference", "equation")
                    a} else{
                      if (input$Function_choose=="fba"){
                        colnames(a) <- c("reaction_id", "reaction_flux", "equation", "gene_associations")
                        a} else {
                          if (input$Function_choose=="fva"){
                            colnames(a) <- c("reaction_id", "lowest_reaction_flux", "highest_reaction_flux", "equation")
                            a
                          }
                          
                        }  
                    }
                }
            }
        }
      }
  })
}
    
        

# step 3: run shiny
shinyApp(ui, server)


